import subprocess
import sys
import os
import simpleworkspace.loader as sw
import toml
import unittest 
from simpleworkspace.utility.module import ModuleInfo
from simpleworkspace.utility.time import StopWatch
from typing import Callable
from simpleworkspace.logproviders import StdoutLogger
from logging import Logger
from simpleworkspace.utility import strings as _strings

class SetupToolsBundler:
    def __init__(self, logger:Logger = None):
        mainModule = ModuleInfo(sys.modules['__main__'])
        self.entryPath = mainModule.pathInfo.Parent.AbsolutePath
        self.logger = StdoutLogger.GetNormalLogger() if logger is None else logger
        self.stopwatch = StopWatch()

    def Command(self, args:list[str], title=None):
        if(title is None):
            title = f'{args}'

        self.logger.info(f"Executing command {title}...")
        stopwatch = StopWatch()
        stopwatch.Start()
        result = subprocess.run(args, text=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if(result.returncode != 0): #something went bad
            raise RuntimeError(f"command failed... stdout: {result.stdout}; stderr: {result.stderr};")
        stdout, stderr = "", ""
        if(result.stdout):
            stdout = "\nSTDOUT: \n" + _strings.IndentText(result.stdout, indentLevel=1, indentStyle='    ') + ";\n"
        if(result.stderr):
            stdout = "\nSTDERR: \n" + _strings.IndentText(result.stderr, indentLevel=1, indentStyle='    ') + ";\n"
        return self.logger.info(f'{stdout}{stderr} - Command finished in {stopwatch.GetElapsedSeconds(2)} seconds...')

    def _CleanUp(self):
        if(os.path.isdir(f'{self.entryPath}/dist')):
            sw.io.directory.RemoveTree(f'{self.entryPath}/dist')

    def Pipe_Init(self):
        self.stopwatch.Start()
        self._CleanUp() #clean leftovers from previous runs

    def Pipe_RunTests(self, testpath='tests/'):
        ### unittests ###
        self.logger.info("Running unittests...")
        test_loader = unittest.TestLoader()
        test_suite = test_loader.discover(os.path.join(self.entryPath, testpath))
        test_runner = unittest.TextTestRunner(verbosity=2)
        result = test_runner.run(test_suite)
        if(result.failures or result.errors): #something went bad
            raise Exception("Unittests failed!")

    def Pipe_IncrementPackageVersion(self):
        ### increment module version ###
        pyProjectData = toml.load("./pyproject.toml")
        oldVersion = pyProjectData["project"]["version"]
        versionInfo = pyProjectData["project"]["version"].split(".")
        versionInfo[2] = str(int(versionInfo[2]) + 1)
        newVersion = ".".join(versionInfo)
        pyProjectData["project"]["version"] = newVersion
        sw.io.file.Create("./pyproject.toml", toml.dumps(pyProjectData))
        self.logger.info(f"Incremented package version from {oldVersion} -> {newVersion}...")

    def Pipe_Build(self):
        ### build package ###
        self.Command([sys.executable, '-m', 'build', '.'])

    def Pipe_Install(self, developmentMode=False):
        ### install on computer as editable/dev mode ###
        if(developmentMode):
            self.Command([sys.executable, "-m", "pip", "install", "--editable", "."])
        else:
            self.Command([sys.executable, "-m", "pip", "install", "."])

    def Pipe_Publish(self, username:str, token:str):
        ### upload to pypi ###
        self.Command(
            [sys.executable, "-m",
                "twine", "upload",
                "-u", username, 
                "-p", token,
                f"{self.entryPath}/dist/*"
            ], 
            title='Upload To PyPi')

    def Pipe_Finish(self):
        self._CleanUp()
        self.logger.info(f"Installer finished! Elapsed: {self.stopwatch.GetElapsedSeconds(decimalPrecision=1)} seconds")


